/*
 * SPDX-FileCopyrightText: 2023 jordi fita mas <jfita@peritasoft.com>
 * SPDX-License-Identifier: AGPL-3.0-only
 */

package types

import (
	"context"
	"net/http"

	"github.com/jackc/pgx/v4"

	"dev.tandem.ws/tandem/camper/pkg/auth"
	"dev.tandem.ws/tandem/camper/pkg/database"
	"dev.tandem.ws/tandem/camper/pkg/form"
	httplib "dev.tandem.ws/tandem/camper/pkg/http"
	"dev.tandem.ws/tandem/camper/pkg/locale"
	"dev.tandem.ws/tandem/camper/pkg/template"
	"dev.tandem.ws/tandem/camper/pkg/uuid"
)

type AdminHandler struct {
	locales locale.Locales
}

func NewAdminHandler(locales locale.Locales) *AdminHandler {
	return &AdminHandler{locales}
}

func (h *AdminHandler) Handler(user *auth.User, company *auth.Company, conn *database.Conn) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		var head string
		head, r.URL.Path = httplib.ShiftPath(r.URL.Path)

		switch head {
		case "new":
			switch r.Method {
			case http.MethodGet:
				f := newTypeForm()
				f.MustRender(w, r, user, company)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet)
			}
		case "":
			switch r.Method {
			case http.MethodGet:
				serveTypeIndex(w, r, user, company, conn)
			case http.MethodPost:
				addType(w, r, user, company, conn)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPost)
			}
		default:
			if !uuid.Valid(head) {
				http.NotFound(w, r)
				return
			}
			f := newTypeForm()
			if err := f.FillFromDatabase(r.Context(), conn, head); err != nil {
				if database.ErrorIsNotFound(err) {
					http.NotFound(w, r)
					return
				}
				panic(err)
			}
			h.typeHandler(user, company, conn, f).ServeHTTP(w, r)
		}
	})
}

func (h *AdminHandler) typeHandler(user *auth.User, company *auth.Company, conn *database.Conn, f *typeForm) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		var head string
		head, r.URL.Path = httplib.ShiftPath(r.URL.Path)

		switch head {
		case "":
			switch r.Method {
			case http.MethodGet:
				f.MustRender(w, r, user, company)
			case http.MethodPut:
				editType(w, r, user, company, conn, f)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPut)
			}
		default:
			loc, ok := h.locales.Get(head)
			if !ok {
				http.NotFound(w, r)
				return
			}
			l10n := newTypeL10nForm(f, loc)
			if err := l10n.FillFromDatabase(r.Context(), conn); err != nil {
				panic(err)
			}
			switch r.Method {
			case http.MethodGet:
				l10n.MustRender(w, r, user, company)
			case http.MethodPut:
				editTypeL10n(w, r, user, company, conn, l10n)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPut)
			}
		}
	})
}

func serveTypeIndex(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn) {
	types, err := collectTypeEntries(r.Context(), company, conn)
	if err != nil {
		panic(err)
	}
	page := &typeIndex{
		Types: types,
	}
	page.MustRender(w, r, user, company)
}

type typeEntry struct {
	Slug         string
	Name         string
	Active       bool
	Translations []*translation
}

type translation struct {
	Language string
	Endonym  string
	Missing  bool
}

func collectTypeEntries(ctx context.Context, company *auth.Company, conn *database.Conn) ([]*typeEntry, error) {
	rows, err := conn.Query(ctx, `
		select campsite_type.slug
		     , campsite_type.name
		     , campsite_type.active
		     , array_agg((lang_tag, endonym, not exists (select 1 from campsite_type_i18n as i18n where i18n.campsite_type_id = campsite_type.campsite_type_id and i18n.lang_tag = language.lang_tag)) order by endonym)
		from campsite_type
		    join company using (company_id)
		   , language
		where lang_tag <> default_lang_tag
		  and language.selectable
		  and campsite_type.company_id = $1
		group by campsite_type.slug
		       , campsite_type.name
		       , campsite_type.active
		order by name
	`, pgx.QueryResultFormats{pgx.BinaryFormatCode}, company.ID)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	var types []*typeEntry
	for rows.Next() {
		entry := &typeEntry{}
		var translations database.RecordArray
		if err = rows.Scan(&entry.Slug, &entry.Name, &entry.Active, &translations); err != nil {
			return nil, err
		}
		for _, el := range translations.Elements {
			entry.Translations = append(entry.Translations, &translation{
				el.Fields[0].Get().(string),
				el.Fields[1].Get().(string),
				el.Fields[2].Get().(bool),
			})
		}
		types = append(types, entry)
	}

	return types, nil
}

type typeIndex struct {
	Types []*typeEntry
}

func (page *typeIndex) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/type/index.gohtml", page)
}

func addType(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn) {
	f := newTypeForm()
	processTypeForm(w, r, user, company, conn, f, func(ctx context.Context) {
		conn.MustExec(ctx, "select add_campsite_type($1, $2, $3, $4)", company.ID, f.Media, f.Name, f.Description)
	})
}

func editType(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, f *typeForm) {
	processTypeForm(w, r, user, company, conn, f, func(ctx context.Context) {
		conn.MustExec(ctx, "select edit_campsite_type($1, $2, $3, $4, $5)", f.Slug, f.Media, f.Name, f.Description, f.Active)
	})
}

func processTypeForm(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, f *typeForm, act func(ctx context.Context)) {
	if err := f.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := user.VerifyCSRFToken(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if ok, err := f.Valid(r.Context(), conn, user.Locale); err != nil {
		panic(err)
	} else if !ok {
		if !httplib.IsHTMxRequest(r) {
			w.WriteHeader(http.StatusUnprocessableEntity)
		}
		f.MustRender(w, r, user, company)
		return
	}
	act(r.Context())
	httplib.Redirect(w, r, "/admin/campsites/types", http.StatusSeeOther)
}

type typeForm struct {
	Slug        string
	Active      *form.Checkbox
	Media       *form.Media
	Name        *form.Input
	Description *form.Input
}

func newTypeForm() *typeForm {
	return &typeForm{
		Active: &form.Checkbox{
			Name:    "active",
			Checked: true,
		},
		Media: &form.Media{
			Input: &form.Input{
				Name: "media",
			},
			Label:  locale.PgettextNoop("Cover image", "input"),
			Prompt: locale.PgettextNoop("Set campsite type cover", "action"),
		},
		Name: &form.Input{
			Name: "name",
		},
		Description: &form.Input{
			Name: "description",
		},
	}
}

func (f *typeForm) FillFromDatabase(ctx context.Context, conn *database.Conn, slug string) error {
	f.Slug = slug
	row := conn.QueryRow(ctx, `
		select name
		     , description
		     , media_id::text
		     , active
		from campsite_type
		where slug = $1
	`, slug)
	return row.Scan(&f.Name.Val, &f.Description.Val, &f.Media.Val, &f.Active.Checked)
}

func (f *typeForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	f.Active.FillValue(r)
	f.Name.FillValue(r)
	f.Description.FillValue(r)
	f.Media.FillValue(r)
	return nil
}

func (f *typeForm) Valid(ctx context.Context, conn *database.Conn, l *locale.Locale) (bool, error) {
	v := form.NewValidator(l)
	v.CheckRequired(f.Name, l.GettextNoop("Name can not be empty."))
	if v.CheckRequired(f.Media.Input, l.GettextNoop("Cover image can not be empty.")) {
		if _, err := v.CheckImageMedia(ctx, conn, f.Media.Input, l.GettextNoop("Cover image must be an image media type.")); err != nil {
			return false, err
		}
	}
	return v.AllOK, nil
}

func (f *typeForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/type/form.gohtml", f)
}
